package mvd.jester;

import java.lang.reflect.Constructor;
import java.time.LocalTime;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import com.google.common.collect.HashBasedTable;
import com.google.common.collect.Table;
import mvd.jester.model.SystemSetup;
import mvd.jester.priority.PriorityManager;
import mvd.jester.simulator.AbstractSimulator;
import mvd.jester.tests.AbstractTest;
import mvd.jester.utils.Logger;

/**
 * TestEnvironment
 */
public class TestEnvironment {

    private final long numberOfTaskSets;
    private final SystemSetup systemSetup;
    private final SystemSetup.Builder builder;
    private final Set<Constructor<? extends AbstractTest>> abstractTests;
    private final Set<Constructor<? extends AbstractSimulator>> abstractSimulators;
    private final Set<PriorityManager> schedulingAlgorithms;

    public TestEnvironment(SystemSetup.Builder builder, long numberOfTaskSets) {
        this.numberOfTaskSets = numberOfTaskSets;
        this.abstractTests = new HashSet<>();
        this.abstractSimulators = new HashSet<>();
        this.schedulingAlgorithms = new HashSet<>();

        this.builder = builder;
        this.systemSetup = builder.build();
    }

    public TestEnvironment registerSchedulingAlgorithm(PriorityManager priorityManager) {
        schedulingAlgorithms.add(priorityManager);

        return this;
    }

    public TestEnvironment registerTest(Class<? extends AbstractTest> abstractTest) {
        try {
            abstractTests.add(abstractTest.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract test!");
        }

        return this;
    }

    public TestEnvironment registerSimulator(Class<? extends AbstractSimulator> abstractSimulator) {
        try {
            abstractSimulators.add(abstractSimulator.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract simulator!");
        }

        return this;
    }

    public TestEnvironment registerTestPair(Class<? extends AbstractTest> abstractTest,
            Class<? extends AbstractSimulator> abstractSimulator) {
        try {
            abstractTests.add(abstractTest.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract test!");
        }
        try {
            abstractSimulators.add(abstractSimulator.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract simulator!");
        }

        return this;
    }

    public void runTests() {
        Set<TestPair> abstractTestInstances = new HashSet<>();
        Set<SimulatorPair> abstractSimulatorInstances = new HashSet<>();

        Map<Long, Long> totalNumberOfTasks = new HashMap<>();

        Table<Long, TestPair, Long> testResults = HashBasedTable.create();
        Table<Long, SimulatorPair, Long> simulatorResults = HashBasedTable.create();

        for (PriorityManager pm : schedulingAlgorithms) {
            for (Constructor<? extends AbstractTest> c : abstractTests) {
                try {
                    abstractTestInstances.add(new TestPair(pm, c.newInstance(this.systemSetup)));
                } catch (Exception e) {
                    System.out.println("Could not instantiate object of AbstractTest!");
                }
            }

            for (Constructor<? extends AbstractSimulator> c : abstractSimulators) {
                try {
                    abstractSimulatorInstances
                            .add(new SimulatorPair(pm, c.newInstance(this.systemSetup)));
                } catch (Exception e) {
                    System.out.println("Could not instantiate object of AbstractSimulator!");
                }
            }
        }

        long checkedTasksets = 0;

        while (checkedTasksets < numberOfTaskSets) {
            builder.rebuild(this.systemSetup);

            double utilization = this.systemSetup.getUtilization();

            while (utilization <= this.systemSetup.getNumberOfProcessors()
                    && checkedTasksets < numberOfTaskSets) {
                checkedTasksets++;

                if (checkedTasksets % 100 == 0) {
                    System.out.println(checkedTasksets + " bereits geschafft");
                }

                long roundedUtilization = (long) (utilization * 10);

                totalNumberOfTasks.compute(roundedUtilization, (k, v) -> v == null ? 1 : v + 1);

                for (TestPair testInstance : abstractTestInstances) {
                    PriorityManager priorityManager = testInstance.getPriorityManager();
                    AbstractTest abstractTest = testInstance.getAbstractTest();
                    if (!testResults.contains(roundedUtilization, testInstance)) {
                        testResults.put(roundedUtilization, testInstance, (long) 0);
                    }

                    if (priorityManager.hasTest(abstractTest)
                            && abstractTest.runSchedulabilityCheck(priorityManager)) {
                        Long val = testResults.get(roundedUtilization, testInstance);
                        testResults.put(roundedUtilization, testInstance, val + 1);
                    }
                }

                for (SimulatorPair simulatorInstance : abstractSimulatorInstances) {
                    PriorityManager priorityManager = simulatorInstance.getPriorityManager();
                    AbstractSimulator abstractSimulator = simulatorInstance.getAbstractSimulator();
                    if (!simulatorResults.contains(roundedUtilization, simulatorInstance)) {
                        simulatorResults.put(roundedUtilization, simulatorInstance, (long) 0);
                    }

                    if (priorityManager.hasSimulator(abstractSimulator)
                            && abstractSimulator.runSimulation(priorityManager)) {
                        Long val = simulatorResults.get(roundedUtilization, simulatorInstance);
                        simulatorResults.put(roundedUtilization, simulatorInstance, val + 1);
                    }
                }

                builder.addTask(systemSetup);

                utilization = this.systemSetup.getUtilization();
            }
        }

        logTestResults(testResults, totalNumberOfTasks);
        logSimulationResults(simulatorResults, totalNumberOfTasks);
    }

    private void logTestResults(Table<Long, TestPair, Long> results,
            Map<Long, Long> totalNumberOfTasks) {
        LocalTime date = LocalTime.now();
        Logger log = new Logger("./results/results_" + systemSetup.getNumberOfProcessors() + "_"
                + date.getHour() + ":" + date.getMinute() + ".txt");
        String firstLine = new String("Utilization");

        if (!results.isEmpty()) {
            for (TestPair testPair : results.columnKeySet()) {
                firstLine = firstLine + "\t" + testPair.getName();
            }

            log.log(firstLine);

            for (Long util : totalNumberOfTasks.keySet()) {
                String line =
                        String.valueOf((double) util / 10) + "\t" + totalNumberOfTasks.get(util);
                for (TestPair simulatorPair : results.columnKeySet()) {
                    line += "\t" + results.get(util, simulatorPair);
                }
                log.log(line);
            }
        }

        log.finalize();
    }

    private void logSimulationResults(Table<Long, SimulatorPair, Long> results,
            Map<Long, Long> totalNumberOfTasks) {
        LocalTime date = LocalTime.now();
        Logger log = new Logger("./results/results_sim_" + systemSetup.getNumberOfProcessors() + "_"
                + date.getHour() + ":" + date.getMinute() + ".txt");
        String firstLine = new String("Utilization\tTotal");

        if (!results.isEmpty()) {
            for (SimulatorPair simulatorPair : results.columnKeySet()) {
                firstLine = firstLine + "\t" + simulatorPair.getName();
            }

            log.log(firstLine);

            for (Long kv : totalNumberOfTasks.keySet()) {
                String line = String.valueOf((double) kv / 10) + "\t" + totalNumberOfTasks.get(kv);
                for (SimulatorPair simulatorPair : results.columnKeySet()) {
                    line += "\t" + results.get(kv, simulatorPair);
                }
                log.log(line);
            }
        }

        log.finalize();
    }


    private class SimulatorPair {
        private final PriorityManager priorityManager;
        private final AbstractSimulator abstractSimulator;


        public SimulatorPair(PriorityManager priorityManager, AbstractSimulator abstractSimulator) {
            this.abstractSimulator = abstractSimulator;
            this.priorityManager = priorityManager;
        }

        /**
         * @return the priorityManager
         */
        public PriorityManager getPriorityManager() {
            return priorityManager;
        }

        /**
         * @return the abstractSimulator
         */
        public AbstractSimulator getAbstractSimulator() {
            return abstractSimulator;
        }

        public String getName() {
            return abstractSimulator.getName() + "_" + priorityManager.getName();
        }
    }

    private class TestPair {
        private final PriorityManager priorityManager;
        private final AbstractTest abstractTest;


        public TestPair(PriorityManager priorityManager, AbstractTest abstractTest) {
            this.abstractTest = abstractTest;
            this.priorityManager = priorityManager;
        }

        /**
         * @return the priorityManager
         */
        public PriorityManager getPriorityManager() {
            return priorityManager;
        }

        /**
         * @return the abstractTest
         */
        public AbstractTest getAbstractTest() {
            return abstractTest;
        }

        public String getName() {
            return abstractTest.getName() + "_" + priorityManager.getName();
        }
    }
}


